# -*- coding: utf-8 -*-
"""
colibri.spec
~~~~~~~~~~~~

Implementation of AMQP specification.

This file was generated 2014-09-15 15:42:09.973725 from amqp0-9-1.extended.xml.

"""

from __future__ import absolute_import

from struct import pack, unpack

from .serialization import (dumps, loads, BIT, OCTET, SHORT, LONG,
                            LONGLONG, SHORTSTR, LONGSTR, TABLE)


def select_method(method_type, unpack=unpack):
    code = unpack('!HH', method_type)
    return METHODS[code]


class MethodMeta(type):

    def __hash__(cls):
        return hash(cls.method_type)

    def __eq__(cls, other):
        if not isinstance(other, tuple):
            other = other.method_type
        return cls.method_type == other


class BaseMethod(object, metaclass=MethodMeta):
    """Base class for all AMQP methods."""

    @classmethod
    def unpack(cls, raw):
        return loads(''.join(type_ for _, type_ in cls.field_info), raw)[0]

    @classmethod
    def pack(cls, payload):
        payload = list(payload)
        fmt, values = [], []
        for name, type_ in cls.field_info:
            fmt.append(type_)
            values.append(payload.pop(0))
        return pack('!HH', *cls.method_type) + dumps(''.join(fmt), values)


class ConnectionStart(BaseMethod):
    """This method starts the connection negotiation process by telling the
    client the protocol version that the server proposes, along with a list of
    security mechanisms which the client can use for authentication.

    Arguments:
        version_major: OCTET
        version_minor: OCTET
        server_properties: TABLE
        mechanisms: LONGSTR
        locales: LONGSTR
    """
    method_type = (10, 10)
    field_info = (
        ('version_major', OCTET),
        ('version_minor', OCTET),
        ('server_properties', TABLE),
        ('mechanisms', LONGSTR),
        ('locales', LONGSTR),
    )
    synchronous = True


class ConnectionStartOK(BaseMethod):
    """This method selects a SASL security mechanism.

    Arguments:
        client_properties: TABLE
        mechanism: SHORTSTR
        response: LONGSTR
        locale: SHORTSTR
    """
    method_type = (10, 11)
    field_info = (
        ('client_properties', TABLE),
        ('mechanism', SHORTSTR),
        ('response', LONGSTR),
        ('locale', SHORTSTR),
    )
    synchronous = True


class ConnectionSecure(BaseMethod):
    """The SASL protocol works by exchanging challenges and responses until
    both peers have received sufficient information to authenticate each other.
    This method challenges the client to provide more information.

    Arguments:
        challenge: LONGSTR
    """
    method_type = (10, 20)
    field_info = (
        ('challenge', LONGSTR),
    )
    synchronous = True


class ConnectionSecureOK(BaseMethod):
    """This method attempts to authenticate, passing a block of SASL data for
    the security mechanism at the server side.

    Arguments:
        response: LONGSTR
    """
    method_type = (10, 21)
    field_info = (
        ('response', LONGSTR),
    )
    synchronous = True


class ConnectionTune(BaseMethod):
    """This method proposes a set of connection configuration values to the
    client. The client can accept and/or adjust these.

    Arguments:
        channel_max: SHORT
        frame_max: LONG
        heartbeat: SHORT
    """
    method_type = (10, 30)
    field_info = (
        ('channel_max', SHORT),
        ('frame_max', LONG),
        ('heartbeat', SHORT),
    )
    synchronous = True


class ConnectionTuneOK(BaseMethod):
    """This method sends the client's connection tuning parameters to the
    server. Certain fields are negotiated, others provide capability
    information.

    Arguments:
        channel_max: SHORT
        frame_max: LONG
        heartbeat: SHORT
    """
    method_type = (10, 31)
    field_info = (
        ('channel_max', SHORT),
        ('frame_max', LONG),
        ('heartbeat', SHORT),
    )
    synchronous = True


class ConnectionOpen(BaseMethod):
    """This method opens a connection to a virtual host, which is a collection
    of resources, and acts to separate multiple application domains within a
    server. The server may apply arbitrary limits per virtual host, such as the
    number of each type of entity that may be used, per connection and/or in
    total.

    Arguments:
        virtual_host: SHORTSTR
        reserved_1: SHORTSTR
        reserved_2: BIT
    """
    method_type = (10, 40)
    field_info = (
        ('virtual_host', SHORTSTR),
        ('reserved_1', SHORTSTR),
        ('reserved_2', BIT),
    )
    synchronous = True


class ConnectionOpenOK(BaseMethod):
    """This method signals to the client that the connection is ready for use.

    Arguments:
        reserved_1: SHORTSTR
    """
    method_type = (10, 41)
    field_info = (
        ('reserved_1', SHORTSTR),
    )
    synchronous = True


class ConnectionClose(BaseMethod):
    """This method indicates that the sender wants to close the connection.
    This may be due to internal conditions (e.g. a forced shut-down) or due to
    an error handling a specific method, i.e. an exception. When a close is due
    to an exception, the sender provides the class and method id of the method
    which caused the exception.

    Arguments:
        reply_code: SHORT
        reply_text: SHORTSTR
        class_id: SHORT
        method_id: SHORT
    """
    method_type = (10, 50)
    field_info = (
        ('reply_code', SHORT),
        ('reply_text', SHORTSTR),
        ('class_id', SHORT),
        ('method_id', SHORT),
    )
    synchronous = True


class ConnectionCloseOK(BaseMethod):
    """This method confirms a Connection.Close method and tells the recipient
    that it is safe to release resources for the connection and close the
    socket.
    """
    method_type = (10, 51)
    field_info = ()
    synchronous = True


class ChannelOpen(BaseMethod):
    """This method opens a channel to the server.

    Arguments:
        reserved_1: SHORTSTR
    """
    method_type = (20, 10)
    field_info = (
        ('reserved_1', SHORTSTR),
    )
    synchronous = True


class ChannelOpenOK(BaseMethod):
    """This method signals to the client that the channel is ready for use.

    Arguments:
        reserved_1: LONGSTR
    """
    method_type = (20, 11)
    field_info = (
        ('reserved_1', LONGSTR),
    )
    synchronous = True


class ChannelFlow(BaseMethod):
    """This method asks the peer to pause or restart the flow of content data
    sent by a consumer. This is a simple flow-control mechanism that a peer can
    use to avoid overflowing its queues or otherwise finding itself receiving
    more messages than it can process. Note that this method is not intended
    for window control. It does not affect contents returned by Basic.Get-Ok
    methods.

    Arguments:
        active: BIT
    """
    method_type = (20, 20)
    field_info = (
        ('active', BIT),
    )
    synchronous = True


class ChannelFlowOK(BaseMethod):
    """Confirms to the peer that a flow command was received and processed.

    Arguments:
        active: BIT
    """
    method_type = (20, 21)
    field_info = (
        ('active', BIT),
    )
    synchronous = False


class ChannelClose(BaseMethod):
    """This method indicates that the sender wants to close the channel. This
    may be due to internal conditions (e.g. a forced shut-down) or due to an
    error handling a specific method, i.e. an exception. When a close is due to
    an exception, the sender provides the class and method id of the method
    which caused the exception.

    Arguments:
        reply_code: SHORT
        reply_text: SHORTSTR
        class_id: SHORT
        method_id: SHORT
    """
    method_type = (20, 40)
    field_info = (
        ('reply_code', SHORT),
        ('reply_text', SHORTSTR),
        ('class_id', SHORT),
        ('method_id', SHORT),
    )
    synchronous = True


class ChannelCloseOK(BaseMethod):
    """This method confirms a Channel.Close method and tells the recipient
    that it is safe to release resources for the channel.
    """
    method_type = (20, 41)
    field_info = ()
    synchronous = True


class ExchangeDeclare(BaseMethod):
    """This method creates an exchange if it does not already exist, and if
    the exchange exists, verifies that it is of the correct and expected class.

    Arguments:
        reserved_1: SHORT
        exchange: SHORTSTR
        type: SHORTSTR
        passive: BIT
        durable: BIT
        auto_delete: BIT
        internal: BIT
        no_wait: BIT
        arguments: TABLE
    """
    method_type = (40, 10)
    field_info = (
        ('reserved_1', SHORT),
        ('exchange', SHORTSTR),
        ('type', SHORTSTR),
        ('passive', BIT),
        ('durable', BIT),
        ('auto_delete', BIT),
        ('internal', BIT),
        ('no_wait', BIT),
        ('arguments', TABLE),
    )
    synchronous = True


class ExchangeDeclareOK(BaseMethod):
    """This method confirms a Declare method and confirms the name of the
    exchange, essential for automatically-named exchanges.
    """
    method_type = (40, 11)
    field_info = ()
    synchronous = True


class ExchangeDelete(BaseMethod):
    """This method deletes an exchange. When an exchange is deleted all queue
    bindings on the exchange are cancelled.

    Arguments:
        reserved_1: SHORT
        exchange: SHORTSTR
        if_unused: BIT
        no_wait: BIT
    """
    method_type = (40, 20)
    field_info = (
        ('reserved_1', SHORT),
        ('exchange', SHORTSTR),
        ('if_unused', BIT),
        ('no_wait', BIT),
    )
    synchronous = True


class ExchangeDeleteOK(BaseMethod):
    """This method confirms the deletion of an exchange.
    """
    method_type = (40, 21)
    field_info = ()
    synchronous = True


class ExchangeBind(BaseMethod):
    """This method binds an exchange to an exchange.

    Arguments:
        reserved_1: SHORT
        destination: SHORTSTR
        source: SHORTSTR
        routing_key: SHORTSTR
        no_wait: BIT
        arguments: TABLE
    """
    method_type = (40, 30)
    field_info = (
        ('reserved_1', SHORT),
        ('destination', SHORTSTR),
        ('source', SHORTSTR),
        ('routing_key', SHORTSTR),
        ('no_wait', BIT),
        ('arguments', TABLE),
    )
    synchronous = True


class ExchangeBindOK(BaseMethod):
    """This method confirms that the bind was successful.
    """
    method_type = (40, 31)
    field_info = ()
    synchronous = True


class ExchangeUnbind(BaseMethod):
    """This method unbinds an exchange from an exchange.

    Arguments:
        reserved_1: SHORT
        destination: SHORTSTR
        source: SHORTSTR
        routing_key: SHORTSTR
        no_wait: BIT
        arguments: TABLE
    """
    method_type = (40, 40)
    field_info = (
        ('reserved_1', SHORT),
        ('destination', SHORTSTR),
        ('source', SHORTSTR),
        ('routing_key', SHORTSTR),
        ('no_wait', BIT),
        ('arguments', TABLE),
    )
    synchronous = True


class ExchangeUnbindOK(BaseMethod):
    """This method confirms that the unbind was successful.
    """
    method_type = (40, 51)
    field_info = ()
    synchronous = True


class QueueDeclare(BaseMethod):
    """This method creates or checks a queue. When creating a new queue the
    client can specify various properties that control the durability of the
    queue and its contents, and the level of sharing for the queue.

    Arguments:
        reserved_1: SHORT
        queue: SHORTSTR
        passive: BIT
        durable: BIT
        exclusive: BIT
        auto_delete: BIT
        no_wait: BIT
        arguments: TABLE
    """
    method_type = (50, 10)
    field_info = (
        ('reserved_1', SHORT),
        ('queue', SHORTSTR),
        ('passive', BIT),
        ('durable', BIT),
        ('exclusive', BIT),
        ('auto_delete', BIT),
        ('no_wait', BIT),
        ('arguments', TABLE),
    )
    synchronous = True


class QueueDeclareOK(BaseMethod):
    """This method confirms a Declare method and confirms the name of the
    queue, essential for automatically-named queues.

    Arguments:
        queue: SHORTSTR
        message_count: LONG
        consumer_count: LONG
    """
    method_type = (50, 11)
    field_info = (
        ('queue', SHORTSTR),
        ('message_count', LONG),
        ('consumer_count', LONG),
    )
    synchronous = True


class QueueBind(BaseMethod):
    """This method binds a queue to an exchange. Until a queue is bound it
    will not receive any messages. In a classic messaging model,
    store-and-forward queues are bound to a direct exchange and subscription
    queues are bound to a topic exchange.

    Arguments:
        reserved_1: SHORT
        queue: SHORTSTR
        exchange: SHORTSTR
        routing_key: SHORTSTR
        no_wait: BIT
        arguments: TABLE
    """
    method_type = (50, 20)
    field_info = (
        ('reserved_1', SHORT),
        ('queue', SHORTSTR),
        ('exchange', SHORTSTR),
        ('routing_key', SHORTSTR),
        ('no_wait', BIT),
        ('arguments', TABLE),
    )
    synchronous = True


class QueueBindOK(BaseMethod):
    """This method confirms that the bind was successful.
    """
    method_type = (50, 21)
    field_info = ()
    synchronous = True


class QueueUnbind(BaseMethod):
    """This method unbinds a queue from an exchange.

    Arguments:
        reserved_1: SHORT
        queue: SHORTSTR
        exchange: SHORTSTR
        routing_key: SHORTSTR
        arguments: TABLE
    """
    method_type = (50, 50)
    field_info = (
        ('reserved_1', SHORT),
        ('queue', SHORTSTR),
        ('exchange', SHORTSTR),
        ('routing_key', SHORTSTR),
        ('arguments', TABLE),
    )
    synchronous = True


class QueueUnbindOK(BaseMethod):
    """This method confirms that the unbind was successful.
    """
    method_type = (50, 51)
    field_info = ()
    synchronous = True


class QueuePurge(BaseMethod):
    """This method removes all messages from a queue which are not awaiting
    acknowledgment.

    Arguments:
        reserved_1: SHORT
        queue: SHORTSTR
        no_wait: BIT
    """
    method_type = (50, 30)
    field_info = (
        ('reserved_1', SHORT),
        ('queue', SHORTSTR),
        ('no_wait', BIT),
    )
    synchronous = True


class QueuePurgeOK(BaseMethod):
    """This method confirms the purge of a queue.

    Arguments:
        message_count: LONG
    """
    method_type = (50, 31)
    field_info = (
        ('message_count', LONG),
    )
    synchronous = True


class QueueDelete(BaseMethod):
    """This method deletes a queue. When a queue is deleted any pending
    messages are sent to a dead-letter queue if this is defined in the server
    configuration, and all consumers on the queue are cancelled.

    Arguments:
        reserved_1: SHORT
        queue: SHORTSTR
        if_unused: BIT
        if_empty: BIT
        no_wait: BIT
    """
    method_type = (50, 40)
    field_info = (
        ('reserved_1', SHORT),
        ('queue', SHORTSTR),
        ('if_unused', BIT),
        ('if_empty', BIT),
        ('no_wait', BIT),
    )
    synchronous = True


class QueueDeleteOK(BaseMethod):
    """This method confirms the deletion of a queue.

    Arguments:
        message_count: LONG
    """
    method_type = (50, 41)
    field_info = (
        ('message_count', LONG),
    )
    synchronous = True


class BasicQos(BaseMethod):
    """This method requests a specific quality of service. The QoS can be
    specified for the current channel or for all channels on the connection.
    The particular properties and semantics of a qos method always depend on
    the content class semantics. Though the qos method could in principle apply
    to both peers, it is currently meaningful only for the server.

    Arguments:
        prefetch_size: LONG
        prefetch_count: SHORT
        global: BIT
    """
    method_type = (60, 10)
    field_info = (
        ('prefetch_size', LONG),
        ('prefetch_count', SHORT),
        ('global', BIT),
    )
    synchronous = True


class BasicQosOK(BaseMethod):
    """This method tells the client that the requested QoS levels could be
    handled by the server. The requested QoS applies to all active consumers
    until a new QoS is defined.
    """
    method_type = (60, 11)
    field_info = ()
    synchronous = True


class BasicConsume(BaseMethod):
    """This method asks the server to start a "consumer", which is a transient
    request for messages from a specific queue. Consumers last as long as the
    channel they were declared on, or until the client cancels them.

    Arguments:
        reserved_1: SHORT
        queue: SHORTSTR
        consumer_tag: SHORTSTR
        no_local: BIT
        no_ack: BIT
        exclusive: BIT
        no_wait: BIT
        arguments: TABLE
    """
    method_type = (60, 20)
    field_info = (
        ('reserved_1', SHORT),
        ('queue', SHORTSTR),
        ('consumer_tag', SHORTSTR),
        ('no_local', BIT),
        ('no_ack', BIT),
        ('exclusive', BIT),
        ('no_wait', BIT),
        ('arguments', TABLE),
    )
    synchronous = True


class BasicConsumeOK(BaseMethod):
    """The server provides the client with a consumer tag, which is used by
    the client for methods called on the consumer at a later stage.

    Arguments:
        consumer_tag: SHORTSTR
    """
    method_type = (60, 21)
    field_info = (
        ('consumer_tag', SHORTSTR),
    )
    synchronous = True


class BasicCancel(BaseMethod):
    """This method cancels a consumer. This does not affect already delivered
    messages, but it does mean the server will not send any more messages for
    that consumer. The client may receive an arbitrary number of messages in
    between sending the cancel method and receiving the cancel-ok reply. It may
    also be sent from the server to the client in the event of the consumer
    being unexpectedly cancelled (i.e. cancelled for any reason other than the
    server receiving the corresponding basic.cancel from the client). This
    allows clients to be notified of the loss of consumers due to events such
    as queue deletion. Note that as it is not a MUST for clients to accept this
    method from the client, it is advisable for the broker to be able to
    identify those clients that are capable of accepting the method, through
    some means of capability negotiation.

    Arguments:
        consumer_tag: SHORTSTR
        no_wait: BIT
    """
    method_type = (60, 30)
    field_info = (
        ('consumer_tag', SHORTSTR),
        ('no_wait', BIT),
    )
    synchronous = True


class BasicCancelOK(BaseMethod):
    """This method confirms that the cancellation was completed.

    Arguments:
        consumer_tag: SHORTSTR
    """
    method_type = (60, 31)
    field_info = (
        ('consumer_tag', SHORTSTR),
    )
    synchronous = True


class BasicPublish(BaseMethod):
    """This method publishes a message to a specific exchange. The message
    will be routed to queues as defined by the exchange configuration and
    distributed to any active consumers when the transaction, if any, is
    committed.

    Arguments:
        reserved_1: SHORT
        exchange: SHORTSTR
        routing_key: SHORTSTR
        mandatory: BIT
        immediate: BIT
    """
    method_type = (60, 40)
    field_info = (
        ('reserved_1', SHORT),
        ('exchange', SHORTSTR),
        ('routing_key', SHORTSTR),
        ('mandatory', BIT),
        ('immediate', BIT),
    )
    synchronous = False


class BasicReturn(BaseMethod):
    """This method returns an undeliverable message that was published with
    the "immediate" flag set, or an unroutable message published with the
    "mandatory" flag set. The reply code and text provide information about the
    reason that the message was undeliverable.

    Arguments:
        reply_code: SHORT
        reply_text: SHORTSTR
        exchange: SHORTSTR
        routing_key: SHORTSTR
    """
    method_type = (60, 50)
    field_info = (
        ('reply_code', SHORT),
        ('reply_text', SHORTSTR),
        ('exchange', SHORTSTR),
        ('routing_key', SHORTSTR),
    )
    synchronous = False


class BasicDeliver(BaseMethod):
    """This method delivers a message to the client, via a consumer. In the
    asynchronous message delivery model, the client starts a consumer using the
    Consume method, then the server responds with Deliver methods as and when
    messages arrive for that consumer.

    Arguments:
        consumer_tag: SHORTSTR
        delivery_tag: LONGLONG
        redelivered: BIT
        exchange: SHORTSTR
        routing_key: SHORTSTR
    """
    method_type = (60, 60)
    field_info = (
        ('consumer_tag', SHORTSTR),
        ('delivery_tag', LONGLONG),
        ('redelivered', BIT),
        ('exchange', SHORTSTR),
        ('routing_key', SHORTSTR),
    )
    synchronous = False


class BasicGet(BaseMethod):
    """This method provides a direct access to the messages in a queue using a
    synchronous dialogue that is designed for specific types of application
    where synchronous functionality is more important than performance.

    Arguments:
        reserved_1: SHORT
        queue: SHORTSTR
        no_ack: BIT
    """
    method_type = (60, 70)
    field_info = (
        ('reserved_1', SHORT),
        ('queue', SHORTSTR),
        ('no_ack', BIT),
    )
    synchronous = True


class BasicGetOK(BaseMethod):
    """This method delivers a message to the client following a get method. A
    message delivered by 'get-ok' must be acknowledged unless the no-ack option
    was set in the get method.

    Arguments:
        delivery_tag: LONGLONG
        redelivered: BIT
        exchange: SHORTSTR
        routing_key: SHORTSTR
        message_count: LONG
    """
    method_type = (60, 71)
    field_info = (
        ('delivery_tag', LONGLONG),
        ('redelivered', BIT),
        ('exchange', SHORTSTR),
        ('routing_key', SHORTSTR),
        ('message_count', LONG),
    )
    synchronous = True


class BasicGetEmpty(BaseMethod):
    """This method tells the client that the queue has no messages available
    for the client.

    Arguments:
        reserved_1: SHORTSTR
    """
    method_type = (60, 72)
    field_info = (
        ('reserved_1', SHORTSTR),
    )
    synchronous = True


class BasicAck(BaseMethod):
    """When sent by the client, this method acknowledges one or more messages
    delivered via the Deliver or Get-Ok methods. When sent by server, this
    method acknowledges one or more messages published with the Publish method
    on a channel in confirm mode. The acknowledgement can be for a single
    message or a set of messages up to and including a specific message.

    Arguments:
        delivery_tag: LONGLONG
        multiple: BIT
    """
    method_type = (60, 80)
    field_info = (
        ('delivery_tag', LONGLONG),
        ('multiple', BIT),
    )
    synchronous = False


class BasicReject(BaseMethod):
    """This method allows a client to reject a message. It can be used to
    interrupt and cancel large incoming messages, or return untreatable
    messages to their original queue.

    Arguments:
        delivery_tag: LONGLONG
        requeue: BIT
    """
    method_type = (60, 90)
    field_info = (
        ('delivery_tag', LONGLONG),
        ('requeue', BIT),
    )
    synchronous = False


class BasicRecoverAsync(BaseMethod):
    """This method asks the server to redeliver all unacknowledged messages on
    a specified channel. Zero or more messages may be redelivered. This method
    is deprecated in favour of the synchronous Recover/Recover-Ok.

    Arguments:
        requeue: BIT
    """
    method_type = (60, 100)
    field_info = (
        ('requeue', BIT),
    )
    synchronous = False


class BasicRecover(BaseMethod):
    """This method asks the server to redeliver all unacknowledged messages on
    a specified channel. Zero or more messages may be redelivered. This method
    replaces the asynchronous Recover.

    Arguments:
        requeue: BIT
    """
    method_type = (60, 110)
    field_info = (
        ('requeue', BIT),
    )
    synchronous = False


class BasicRecoverOK(BaseMethod):
    """This method acknowledges a Basic.Recover method.
    """
    method_type = (60, 111)
    field_info = ()
    synchronous = True


class BasicNack(BaseMethod):
    """This method allows a client to reject one or more incoming messages. It
    can be used to interrupt and cancel large incoming messages, or return
    untreatable messages to their original queue. This method is also used by
    the server to inform publishers on channels in confirm mode of unhandled
    messages. If a publisher receives this method, it probably needs to
    republish the offending messages.

    Arguments:
        delivery_tag: LONGLONG
        multiple: BIT
        requeue: BIT
    """
    method_type = (60, 120)
    field_info = (
        ('delivery_tag', LONGLONG),
        ('multiple', BIT),
        ('requeue', BIT),
    )
    synchronous = False


class TxSelect(BaseMethod):
    """This method sets the channel to use standard transactions. The client
    must use this method at least once on a channel before using the Commit or
    Rollback methods.
    """
    method_type = (90, 10)
    field_info = ()
    synchronous = True


class TxSelectOK(BaseMethod):
    """This method confirms to the client that the channel was successfully
    set to use standard transactions.
    """
    method_type = (90, 11)
    field_info = ()
    synchronous = True


class TxCommit(BaseMethod):
    """This method commits all message publications and acknowledgments
    performed in the current transaction. A new transaction starts immediately
    after a commit.
    """
    method_type = (90, 20)
    field_info = ()
    synchronous = True


class TxCommitOK(BaseMethod):
    """This method confirms to the client that the commit succeeded. Note that
    if a commit fails, the server raises a channel exception.
    """
    method_type = (90, 21)
    field_info = ()
    synchronous = True


class TxRollback(BaseMethod):
    """This method abandons all message publications and acknowledgments
    performed in the current transaction. A new transaction starts immediately
    after a rollback. Note that unacked messages will not be automatically
    redelivered by rollback; if that is required an explicit recover call
    should be issued.
    """
    method_type = (90, 30)
    field_info = ()
    synchronous = True


class TxRollbackOK(BaseMethod):
    """This method confirms to the client that the rollback succeeded. Note
    that if an rollback fails, the server raises a channel exception.
    """
    method_type = (90, 31)
    field_info = ()
    synchronous = True


class ConfirmSelect(BaseMethod):
    """This method sets the channel to use publisher acknowledgements. The
    client can only use this method on a non-transactional channel.

    Arguments:
        nowait: BIT
    """
    method_type = (85, 10)
    field_info = (
        ('nowait', BIT),
    )
    synchronous = True


class ConfirmSelectOK(BaseMethod):
    """This method confirms to the client that the channel was successfully
    set to use publisher acknowledgements.
    """
    method_type = (85, 11)
    field_info = ()
    synchronous = True


METHODS = {
    (10, 10): ConnectionStart,
    (10, 11): ConnectionStartOK,
    (10, 20): ConnectionSecure,
    (10, 21): ConnectionSecureOK,
    (10, 30): ConnectionTune,
    (10, 31): ConnectionTuneOK,
    (10, 40): ConnectionOpen,
    (10, 41): ConnectionOpenOK,
    (10, 50): ConnectionClose,
    (10, 51): ConnectionCloseOK,
    (20, 10): ChannelOpen,
    (20, 11): ChannelOpenOK,
    (20, 20): ChannelFlow,
    (20, 21): ChannelFlowOK,
    (20, 40): ChannelClose,
    (20, 41): ChannelCloseOK,
    (40, 10): ExchangeDeclare,
    (40, 11): ExchangeDeclareOK,
    (40, 20): ExchangeDelete,
    (40, 21): ExchangeDeleteOK,
    (40, 30): ExchangeBind,
    (40, 31): ExchangeBindOK,
    (40, 40): ExchangeUnbind,
    (40, 51): ExchangeUnbindOK,
    (50, 10): QueueDeclare,
    (50, 11): QueueDeclareOK,
    (50, 20): QueueBind,
    (50, 21): QueueBindOK,
    (50, 50): QueueUnbind,
    (50, 51): QueueUnbindOK,
    (50, 30): QueuePurge,
    (50, 31): QueuePurgeOK,
    (50, 40): QueueDelete,
    (50, 41): QueueDeleteOK,
    (60, 10): BasicQos,
    (60, 11): BasicQosOK,
    (60, 20): BasicConsume,
    (60, 21): BasicConsumeOK,
    (60, 30): BasicCancel,
    (60, 31): BasicCancelOK,
    (60, 40): BasicPublish,
    (60, 50): BasicReturn,
    (60, 60): BasicDeliver,
    (60, 70): BasicGet,
    (60, 71): BasicGetOK,
    (60, 72): BasicGetEmpty,
    (60, 80): BasicAck,
    (60, 90): BasicReject,
    (60, 100): BasicRecoverAsync,
    (60, 110): BasicRecover,
    (60, 111): BasicRecoverOK,
    (60, 120): BasicNack,
    (90, 10): TxSelect,
    (90, 11): TxSelectOK,
    (90, 20): TxCommit,
    (90, 21): TxCommitOK,
    (90, 30): TxRollback,
    (90, 31): TxRollbackOK,
    (85, 10): ConfirmSelect,
    (85, 11): ConfirmSelectOK,
}

# AMQP constants
FRAME_METHOD = 1
FRAME_HEADER = 2
FRAME_BODY = 3
FRAME_HEARTBEAT = 8
FRAME_MIN_SIZE = 4096
FRAME_END = 206
REPLY_SUCCESS = 200
CONTENT_TOO_LARGE = 311
NO_CONSUMERS = 313
CONNECTION_FORCED = 320
INVALID_PATH = 402
ACCESS_REFUSED = 403
NOT_FOUND = 404
RESOURCE_LOCKED = 405
PRECONDITION_FAILED = 406
FRAME_ERROR = 501
SYNTAX_ERROR = 502
COMMAND_INVALID = 503
CHANNEL_ERROR = 504
UNEXPECTED_FRAME = 505
RESOURCE_ERROR = 506
NOT_ALLOWED = 530
NOT_IMPLEMENTED = 540
INTERNAL_ERROR = 541


# -*- coding: utf-8 -*-
from __future__ import absolute_import

from ._version import __version__  # noqa

from .message import BasicMessage
from .connection import Connection

__all__ = ['Connection', 'BasicMessage']



# -*- coding: utf-8 -*-
"""
colibri.frame
~~~~~~~~~~~~~

This module contains classes that represent different kinds of AMQP
frames.

"""
from contextlib import contextmanager

from asyncio import Lock, Future, async

from .serialization import dumps
from .message import load_properties, BasicMessage
from .spec import (select_method, FRAME_METHOD, FRAME_HEADER,
                   FRAME_BODY, FRAME_HEARTBEAT, FRAME_END, METHODS,
                   BasicReturn, BasicDeliver, BasicGetOK)
from .exceptions import UnexpectedFrame

__all__ = ['MethodFrame', 'HeaderFrame', 'BodyFrame',
           'HeartbeatFrame', 'FrameHandler', 'FRAME_BY_TYPE']


def pack(frame_type, channel_id, payload, dumps=dumps, frame_end=FRAME_END):
    return (dumps('oBl', (frame_type, channel_id, len(payload))) +
            payload + dumps('o', FRAME_END))


class FrameMeta(type):

    def __eq__(cls, other):
        return cls.frame_type == other.frame_type

    def __neq__(cls, other):
        return cls.frame_type != other.frame_type


class MethodFrame(object, metaclass=FrameMeta):

    frame_type = FRAME_METHOD

    @classmethod
    def pack(cls, channel_id, method, payload):
        return pack(cls.frame_type, channel_id, method.pack(payload))

    @classmethod
    def unpack(cls, payload, select_method=select_method):
        method_type, payload = payload[:4], payload[4:]
        method = select_method(method_type)
        return method.unpack(payload), method


class HeaderFrame(object, metaclass=FrameMeta):

    frame_type = FRAME_HEADER

    properties = BasicMessage.PROPERTIES

    @classmethod
    def pack(cls, channel_id, payload, pack=pack):
        return pack(cls.frame_type, channel_id, payload)

    @classmethod
    def unpack(cls, payload, load_properties=load_properties):
        return load_properties(payload, cls.properties)


class BodyFrame(object, metaclass=FrameMeta):

    frame_type = FRAME_BODY

    @classmethod
    def pack(cls, channel_id, payload, pack=pack):
        return pack(cls.frame_type, channel_id, payload)

    @classmethod
    def unpack(cls, payload):
        return payload


class HeartbeatFrame(object, metaclass=FrameMeta):

    frame_type = FRAME_HEARTBEAT

    @classmethod
    def pack(cls, pack=pack):
        return pack(cls.frame_type, 0, b'')

    @classmethod
    def unpack(cls, payload):
        return payload


FRAME_BY_TYPE = {
    FRAME_METHOD: MethodFrame,
    FRAME_HEADER: HeaderFrame,
    FRAME_BODY: BodyFrame,
    FRAME_HEARTBEAT: HeartbeatFrame,
}


UNEXPECTED_FRAME_FMT = 'Expected {} but received {}'


class FrameHandler(object):

    def __init__(self, channel):
        self.channel = channel
        self.loop = channel.loop

        self.partial_message = None
        self.callbacks = {}

        self.expected_frame = None
        self.expected_methods = ()
        self.response = None

        self.lock = Lock(loop=self.loop)

    def update(self, d):
        self.callbacks.update(d)

    @contextmanager
    def wait_for(self, *expected):
        yield from self.lock.acquire()
        self.expected_frame = MethodFrame
        self.expected_methods = expected
        r = self.response = Future(loop=self.loop)
        # Some methods may not wait for the reply
        # In this case the code below will allow to instantly return
        # when yielding from r
        if expected[0] is None:
            r.set_result(None)
        try:
            yield r
        finally:
            self.expected_frame = None
            self.expected_methods = ()
            self.response = None
            self.lock.release()

    def ready(self, result=None):
        self.response.set_result(result)

    def handle(self, frame, payload,
               CONTENT_SPEC=(BasicReturn, BasicDeliver, BasicGetOK)):
        # At first, check if an unexpected frame was received
        response = self.response
        if self.expected_frame is not None and frame != self.expected_frame:
            exc = UnexpectedFrame(UNEXPECTED_FRAME_FMT.format(
                self.expected_frame, frame
            ))
            response.set_exception(exc)
            return
        if frame == MethodFrame:
            payload, method = frame.unpack(payload)
            if method.synchronous:
                # Check of an unexpected method was received
                if not any(method == cls for cls in self.expected_methods):
                    exc = UnexpectedFrame(UNEXPECTED_FRAME_FMT.format(
                        self.expected_methods, method
                    ))
                    response.set_exception(exc)
                    return
            if any(method == cls for cls in CONTENT_SPEC):
                self.partial_message = BasicMessage()
                self.expected_frame = HeaderFrame
            else:
                try:
                    callback = self.callbacks[method.method_type]
                except KeyError:
                    # Recieved reply and there is no handler for it
                    if method.method_type not in METHODS:
                        raise Exception('Unknown method')
                    self.ready(payload)
                else:
                    # Schedule task to handle the method
                    async(callback(*payload))
        elif frame == HeaderFrame:
            msg = self.partial_message
            msg.add_header(payload)
            if msg._ready:
                # Bodyless message
                self.partial_message = None
                response.set_result(msg)
            else:
                self.expected_frame = BodyFrame
        elif frame == BodyFrame:
            msg = self.partial_message
            msg.add_body_chunk(payload)
            if msg._ready:
                self.partial_message = None
                response.set_result(msg)
                self.expected_frame = None



# -*- coding: utf-8 -*-
"""
colibri.connection
~~~~~~~~~~~~~~~~~~

AMQP connections.

"""
from __future__ import absolute_import

from array import array

from asyncio import coroutine, get_event_loop

from . import __version__, spec
from .protocol import AMQP
from .channel import Channel
from .frame import FrameHandler
from .serialization import dumps
from .utils import cached_property, YieldFromContextManager
from .log import get_logger

log = get_logger(__name__)


LIBRARY_PROPERTIES = {
    'product': 'colibri',
    'product_version': __version__,
    'capabilities': {},
}


class Connection(object):
    """This class represents a connection to AMQP broker"""

    #: Protocol implementation
    AMQP = AMQP

    #: AMQP channel implementation
    Channel = Channel

    def __init__(self, host='localhost', port=5672,
                 userid='guest', password='guest',
                 virtual_host='/', login_method='AMQPLAIN',
                 login_response=None, locale='en_US', client_properties=None,
                 ssl=False, connect_timeout=None, channel_max=None,
                 frame_max=None, heartbeat_interval=0, on_open=None,
                 on_blocked=None, on_unblocked=None, confirm_publish=False,
                 on_tune_ok=None, loop=None, **kwargs):
        channel_max = self.channel_max = channel_max or 65535
        self.frame_max = frame_max or 131072

        if login_response is None:
            if userid is not None and password is not None:
                if isinstance(userid, str):
                    userid = userid.encode('utf-8')
                if isinstance(password, str):
                    password = password.encode('utf-8')
                login_response = dumps('F', {'LOGIN': userid,
                                             'PASSWORD': password})
                # Skip the length at the beginning
                login_response = login_response[4:]

        self.host = host
        self.port = port
        self.client_properties = dict(
            LIBRARY_PROPERTIES, **client_properties or {}
        )
        self.client_properties.setdefault('capabilities', {})

        self.login_method = login_method
        self.login_response = login_response
        self.locale = locale
        self.virtual_host = virtual_host
        self.ssl = ssl

        self.client_heartbeat = heartbeat_interval

        self._available_channel_ids = array('H', range(channel_max, 0, -1))

        self.version_major = 0
        self.version_minor = 0
        self.server_properties = {}
        self.mechanisms = []
        self.locales = []

        self.confirm_publish = confirm_publish

        # Callbacks
        self.on_open = on_open
        self.on_blocked = on_blocked
        self.on_unblocked = on_unblocked

        self.loop = loop or get_event_loop()

        self.channel_id = 0
        self.channels = {0: self}  # self.Channel(self, 0)}

        self.handler = FrameHandler(self)
        self.handle = self.handler.handle

        self._setup_callbacks()

    def send_method(self, method, payload, message=None,
                    frame_size=None, channel_id=None):
        # This method is private and should not be used directly
        if self.protocol is None:
            raise Exception('Call `open` method at first')
        channel_id = channel_id or self.channel_id
        protocol = self.protocol
        # if protocol.closing.done():
        #     raise Exception('Channel #{} is closed'.format(channel_id))
        protocol.send_method(channel_id, method, payload,
                             message=message, frame_size=frame_size)

    @cached_property
    def protocol(self):
        return self.AMQP(self)

    def open(self, host=None, port=None, reply_method=spec.ConnectionStart):
        """Open connection to AMQP broker.

        :keyword host: Override `host` specified in constructor.
        :keyword port: Override `port` specified in constructor.
        """
        host, port = host or self.host, port or self.port
        with self.handler.wait_for(reply_method) as fut:
            yield from self.loop.create_connection(
                lambda: self.protocol, host=host, port=port
            )
            return (yield from fut)

    def __enter__(self):
        raise RuntimeError(
            '"yield from" should be used as context manager exception')

    def __exit__(self, *args):
        pass

    def __iter__(self):
        # This is not a coroutine.  It is meant to enable the idiom:
        #
        #     with (yield from conn):
        #         <block>
        #
        # as an alternative to:
        #
        #     yield from conn.open()
        #     try:
        #         <block>
        #     finally:
        #         conn.close()
        yield from self.open()
        return YieldFromContextManager(self)

    def close(self, reply_code=0, reply_text='', class_id=0, method_id=0,
              method=spec.ConnectionClose,
              reply_method=spec.ConnectionCloseOK):
        """Close connection to AMQP broker.

        :keyword reply_code: The reply code. Must be one of constants defined
            in `colibri.spec`.
        :keyword reply_text: Localized reply text. This text can be logged
            as an aid to resolving issues.
        :keyword class_id: Tell the broker which method (class)
            caused the close.
        :keyword method_id: Tell the broker which method (ID)
            caused the close.
        """
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method, (reply_code, reply_text, class_id, method_id),
            )
            return (yield from fut)

    def channel(self, channel_id=None, on_open=None):
        """Get channel instance. If channel with `channel_id` does not exist,
        it will be created.

        :keyword channel_id: Channel ID. If set to None (default value),
            channel id will be chosen from available channel ids.
        :keyword on_open: Optional callback, must accept a single
            argument - channel instance.
        """
        try:
            return self.channels[channel_id]
        except KeyError:
            chan = self.Channel(self, channel_id, on_open=on_open)
            self.channels[channel_id] = chan
            return chan

    def _get_free_channel_id(self):
        try:
            return self._available_channel_ids.pop()
        except IndexError:
            raise Exception(
                'No free channel ids, current={}, channeld_max={}'.format(
                    len(self.channels), self.channel_max)
            )

    def _claim_channel_id(self, channel_id):
        try:
            return self._available_channel_ids.remove(channel_id)
        except ValueError:
            raise Exception('Channel #{} already open'.format(channel_id))

    def _do_close(self):
        del self.channels[0]
        for channel in self.channels.values():
            channel.close()

        self.protocol.transport.close()
        self.channels = {}

    def _setup_callbacks(self):
        self.handler.update({
            spec.ConnectionStart: self._on_start,
            spec.ConnectionOpenOK: self._on_open_ok,
            spec.ConnectionSecure: self._on_secure,
            spec.ConnectionTune: self._on_tune,
            spec.ConnectionClose: self._on_close,
            # spec.ConnectionBlocked: self._on_blocked,
            # spec.ConnectionUnblocked: self._on_unblocked,
            spec.ConnectionCloseOK: self._on_close_ok,
        })

    @coroutine
    def _on_start(self, version_major, version_minor, server_properties,
                  mechanisms, locales, method=spec.ConnectionStartOK,
                  reply_method=spec.ConnectionTune):
        client_properties = self.client_properties
        self.version_major = version_major
        self.version_minor = version_minor
        self.server_properties = server_properties
        self.mechanisms = mechanisms.split(' ')
        self.locales = locales.split(' ')

        scap = server_properties.get('capabilities') or {}
        if scap.get('consumer_cancel_notify'):
            client_properties['consumer_cancel_notify'] = True
        if scap.get('connection.blocked'):
            client_properties['connection.blocked'] = True

        self.handler.expected_methods = (reply_method,)
        self.send_method(
            method,
            (client_properties, self.login_method,
             self.login_response, locales),
        )

    @coroutine
    def _on_tune(self, channel_max, frame_max, server_heartbeat,
                 method=spec.ConnectionTuneOK):
        cmax = self.channel_max = channel_max or self.channel_max
        fmax = self.frame_max = frame_max or self.frame_max
        sheartbeat = self.server_heartbeat = server_heartbeat or 0
        cheartbeat = self.client_heartbeat

        if sheartbeat == 0 or cheartbeat == 0:
            negotiate = max
        else:
            negotiate = min
        heartbeat = self.heartbeat = negotiate(sheartbeat, cheartbeat)

        # Ignore server heartbeat if client_heartbeat is disabled
        if not cheartbeat:
            self.heartbeat = 0

        self.send_method(method, (cmax, fmax, heartbeat))
        self._after_tune()

    def _after_tune(self, method=spec.ConnectionOpen,
                    reply_method=spec.ConnectionOpenOK):
        self.handler.expected_methods = (reply_method,)
        self.send_method(method, (self.virtual_host, '', False))

    @coroutine
    def _on_open_ok(self, reserved_1):
        self.on_open and self.on_open(self)
        self.handler.ready()

    @coroutine
    def _on_secure(self, challenge):
        print('Challenge is: ', challenge)

    @coroutine
    def _on_close(self, reply_code, reply_text, class_id, method_id,
                  method=spec.ConnectionCloseOK):
        # XXX logging
        self.send_method(method, ())
        self._do_close()
        # XXX raise

    @coroutine
    def _on_close_ok(self):
        self._do_close()
        self.handler.ready()

    @coroutine
    def _on_blocked(self):
        self.on_blocked and self.on_blocked(self)

    @coroutine
    def _on_unblocked(self):
        self.on_unblocked and self.on_unblocked(self)



# -*- coding: utf-8 -*-
"""
colibri.protocol
~~~~~~~~~~~~~~~~

AMQP implementation as `asyncio` Protocol.

"""
from __future__ import absolute_import

from struct import unpack

from asyncio import Protocol, async

from .spec import FRAME_END
from .exceptions import UnexpectedFrame
from .utils import chunks
from .frame import (FRAME_BY_TYPE, MethodFrame, BodyFrame,
                    HeaderFrame, HeartbeatFrame)
from .heartbeat import HeartbeatMonitor


UNEXPECTED_FRAME_FMT = 'Expected {} but received {}'


class AMQP(Protocol):

    HeartbeatMonitor = HeartbeatMonitor

    def __init__(self, connection):
        self.connection = connection
        self.loop = connection.loop

        self.heartbeat = HeartbeatMonitor(protocol=self, interval=10)
        self.prev = bytes()

        self._wait = None
        self._result = None
        self._callbacks = {}

    def connection_made(self, transport):
        """Called once when the connection is established."""
        self.transport = transport
        self.send_protocol_header()

    def connection_lost(self, exc=None):
        """Called once when the connection is lost."""
        if exc is not None:
            # TODO logging
            pass

    def data_received(self, data, bytes=bytes, len=len,
                      unpack=unpack, async=async, FRAME_BY_TYPE=FRAME_BY_TYPE,
                      FRAME_END=FRAME_END):
        """Main callback, called when `data` is received."""
        channel = self.connection.channel
        # channels = self.connection.channels
        maybe_received_heartbeat = self.heartbeat.maybe_received

        while True:
            # Each frame may be a heartbeat frame
            maybe_received_heartbeat()

            data = self.prev + data
            self.prev = bytes()

            len_data = len(data)
            if len_data < 7:
                self.prev = data
                return

            header = data[:7]
            type_, channel_id, size = unpack('!BHL', header)

            if len_data < size + 8:
                self.prev = data
                return

            payload = data[7:7+size]
            ending = data[7+size]

            if ending != FRAME_END:
                self.transport.close()
                raise UnexpectedFrame(UNEXPECTED_FRAME_FMT.format(
                    FRAME_END, ending
                ))

            frame = FRAME_BY_TYPE[type_]
            if frame != HeartbeatFrame:
                channel(channel_id).handle(frame, payload)
                # channels[channel_id].handle(frame, payload)

            reminder = data[8+size:]
            if not reminder:
                return

            # Next loop
            data = reminder

    def send_protocol_header(self, HEADER=b'AMQP\x00\x00\x09\x01'):
        self.transport.write(HEADER)

    def send_method(self, channel_id, method, payload,
                    message=None, frame_size=None):
        frame = MethodFrame.pack(channel_id, method, payload)
        self._send_frame(frame)
        if message is not None:
            assert frame_size is not None
            self._send_message(channel_id, message, frame_size,
                               method.method_type[0])

    def _send_message(self, channel_id, message, frame_size, class_id):
        header_payload = message.dump_properties(class_id)
        header_frame = HeaderFrame.pack(channel_id, header_payload)
        self._send_frame(header_frame)
        for payload in chunks(message.body, frame_size):
            frame = BodyFrame.pack(channel_id, payload)
            self._send_frame(frame)

    def _send_frame(self, frame):
        self.transport.write(frame)



from functools import partial

from asyncio import coroutine

from . import spec
from .frame import FrameHandler
from .message import BasicMessage
from .utils import YieldFromContextManager
from .exceptions import QueueEmpty


class Channel(object):

    def __init__(self, connection, channel_id=None, on_open=None):
        if channel_id:
            connection._claim_channel_id(channel_id)
        else:
            channel_id = connection._get_free_channel_id()

        self.channel_id = channel_id
        self.connection = connection
        self.protocol = connection.protocol
        self.loop = connection.loop

        self.is_open = False
        self.active = True

        self.consumer_callbacks = {}
        self.cancel_callbacks = {}
        self.no_ack_consumers = set()
        self.events = set()

        self._confirm_selected = False

        self.handler = FrameHandler(self)
        self.handle = self.handler.handle

        # In case connection confirms publications
        if connection.confirm_publish:
            self.basic_publish = self.basic_publish_confirm

        self.on_open = on_open
        self.send_method = partial(connection.send_method,
                                   channel_id=channel_id)

        self._setup_callbacks()

    def _setup_callbacks(self):
        self.handler.update({
            spec.ChannelClose: self._on_close,
            spec.ChannelCloseOK: self._on_close_ok,
            spec.ChannelFlow: self._on_flow,
            spec.ChannelOpenOK: self._on_open_ok,
            spec.BasicCancel: self._on_basic_cancel,
            spec.BasicCancelOK: self._on_basic_cancel_ok,
            spec.BasicDeliver: self._on_basic_deliver,
            spec.BasicReturn: self._on_basic_return,
            spec.BasicAck: self._on_basic_ack,
        })

    def open(self):
        return (yield from self._do_open())

    def __enter__(self):
        raise RuntimeError(
            '"yield from" should be used as context manager exception')

    def __exit__(self, *args):
        pass

    def __iter__(self):
        # This is not a coroutine.  It is meant to enable the idiom:
        #
        #     with (yield from chan):
        #         <block>
        #
        # as an alternative to:
        #
        #     yield from chan.open()
        #     try:
        #         <block>
        #     finally:
        #         chan.close()
        yield from self.open()
        return YieldFromContextManager(self)

    @coroutine
    def _on_close(self, reply_code, reply_text, class_id, method_id,
                  method=spec.ChannelCloseOK):
        self.send_method(method, ())
        yield from self.channel._do_revive()
        # raise error_for_code(
        #     reply_code, reply_text, (class_id, method_id), ChannelError,
        # )

    @coroutine
    def _on_close_ok(self):
        self._do_close()
        self.handler.ready()

    @coroutine
    def _on_flow(self, active, method=spec.ChannelFlowOK):
        self.active = active
        self.send_method(method, (active,))

    def _do_open(self, method=spec.ChannelOpen,
                 reply_method=spec.ChannelOpenOK):
        if self.is_open:
            return
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, ('',))
            return (yield from fut)

    @coroutine
    def _on_open_ok(self, reserved_1):
        self.is_open = True
        self.on_open and self.on_open(self)
        # AMQP_LOGGER.debug('Channel open')
        self.handler.ready()

    def _on_basic_cancel(self, consumer_tag):
        callback = self._remove_tag(consumer_tag)
        if callback:
            callback(consumer_tag)
        else:
            print('on basic cancel')
            # raise ConsumerCancelled(consumer_tag, spec.Basic.Cancel)

    def _on_basic_cancel_ok(self, consumer_tag):
        self._remove_tag(consumer_tag)
        self.handler.ready()

    def _remove_tag(self, consumer_tag):
        self.consumer_callbacks.pop(consumer_tag, None)
        self.cancel_callbacks.pop(consumer_tag, None)

    def _on_basic_deliver(self, consumer_tag, delivery_tag, redelivered,
                          exchange, routing_key, msg):
        msg.channel = self
        msg.delivery_info = {
            'consumer_tag': consumer_tag,
            'delivery_tag': delivery_tag,
            'redelivered': redelivered,
            'exchange': exchange,
            'routing_key': routing_key,
        }
        try:
            fun = self.consumer_callbacks[consumer_tag]
        except KeyError:
            pass
        else:
            fun(msg)

    def _on_basic_return(self, reply_code, reply_text,
                         exchange, routing_key, message):
        # exc = error_for_code(
        #     reply_code, reply_text, spec.Basic.Return, ChannelError,
        # )
        # TODO work with events
        exc = None
        handlers = self.events.get('basic_return')
        if not handlers:
            raise exc
        for callback in handlers:
            callback(exc, exchange, routing_key, message)

    def _on_basic_ack(self, delivery_tag, multiple):
        for callback in self.events['basic_ack']:
            callback(delivery_tag, multiple)

    def _do_close(self):
        self.is_open = False
        channel_id, self.channel_id = self.channel_id, None
        connection, self.connection = self.connection, None
        if connection:
            connection.channels.pop(channel_id, None)
            connection._available_channel_ids.append(channel_id)

    def _do_revive(self):
        self.is_open = True
        return (yield from self._do_open())

    def close(self, reply_code=0, reply_text='', class_id=0, method_id=0,
              method=spec.ChannelClose, reply_method=spec.ChannelCloseOK):
        if not self.is_open or self.connection is None:
            return
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (reply_code, reply_text, class_id, method_id),
            )
            res = yield from fut
            self._do_close()
            return res

    def flow(self, active, method=spec.ChannelFlow,
             reply_method=spec.ChannelFlowOK):
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (active,))
            return (yield from fut)

    # Work with Exchanges

    def exchange_declare(self, exchange, type, passive=False, durable=False,
                         auto_delete=False, internal=False, no_wait=False,
                         arguments=None, method=spec.ExchangeDeclare,
                         reply_method=spec.ExchangeDeclareOK):
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, exchange, type, passive, durable, auto_delete,
                 internal, no_wait, arguments),
            )
            return (yield from fut)

    def exchange_delete(self, exchange, if_unused=False, no_wait=False,
                        method=spec.ExchangeDelete,
                        reply_method=spec.ExchangeDeleteOK):
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (0, exchange, if_unused, no_wait))
            return (yield from fut)

    def exchange_bind(self, destination, source='', routing_key='',
                      no_wait=False, arguments=None,
                      method=spec.ExchangeBind,
                      reply_method=spec.ExchangeBindOK):
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, destination, source, routing_key,
                 no_wait, arguments),
            )
            return (yield from fut)

    def exchange_unbind(self, destination, source='', routing_key='',
                        no_wait=False, arguments=None,
                        method=spec.ExchangeUnbind,
                        reply_method=spec.ExchangeUnbindOK):
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, destination, source, routing_key, no_wait, arguments),
            )
            return (yield from fut)

    # Work with Queues

    def queue_bind(self, queue, exchange='', routing_key='', no_wait=False,
                   arguments=None, method=spec.QueueBind,
                   reply_method=spec.QueueBindOK):
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, queue, exchange, routing_key, no_wait, arguments),
            )
            return (yield from fut)

    def queue_unbind(self, queue, exchange='', routing_key='', no_wait=False,
                     arguments=None, method=spec.QueueUnbind,
                     reply_method=spec.QueueUnbindOK):
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, queue, exchange, routing_key, no_wait, arguments),
            )
            return (yield from fut)

    def queue_declare(self, queue, passive=False, durable=False,
                      exclusive=False, auto_delete=True, no_wait=False,
                      arguments=None, method=spec.QueueDeclare,
                      reply_method=spec.QueueDeclareOK):
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, queue, passive, durable, exclusive, auto_delete,
                 no_wait, arguments),
            )
            return (yield from fut)

    def queue_delete(self, queue, if_unused=False, if_empty=False,
                     no_wait=False, method=spec.QueueDelete,
                     reply_method=spec.QueueDeleteOK):
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (0, queue, if_unused, if_empty, no_wait))
            return (yield from fut)

    def queue_purge(self, queue, no_wait=False, method=spec.QueuePurge,
                    reply_method=spec.QueuePurgeOK):
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (0, queue, no_wait))
            return (yield from fut)

    # Work with Basic

    def basic_ack(self, delivery_tag, multiple=False,
                  method=spec.BasicAck):
        self.send_method(method, (delivery_tag, multiple))
        yield

    def basic_cancel(self, consumer_tag, no_wait=False,
                     method=spec.BasicCancel,
                     reply_method=spec.BasicCancelOK):
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (consumer_tag, no_wait))
            return (yield from fut)

    def basic_consume(self, queue, consumer_tag='', no_local=False,
                      no_ack=True, exclusive=False, no_wait=False,
                      arguments=None, consumer_callback=None, on_cancel=None,
                      method=spec.BasicConsume,
                      reply_method=spec.BasicConsumeOK):
        arguments = arguments or {}
        reply_method = None if no_wait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(
                method,
                (0, queue, consumer_tag, no_local, no_ack, exclusive,
                 no_wait, arguments),
            )
            proposed_tag = yield from fut

        # In case consumer_tag was not provided
        if not no_wait and not consumer_tag:
            consumer_tag = proposed_tag

        assert consumer_tag

        self.consumer_callbacks[consumer_tag] = consumer_callback

        if on_cancel:
            self.cancel_callbacks[consumer_tag] = on_cancel
        if no_ack:
            self.no_ack_consumers.add(consumer_tag)
        return consumer_tag

    def basic_get(self, queue, no_ack=True, method=spec.BasicGet,
                  reply_method=(spec.BasicGetOK, spec.BasicGetEmpty)):
        with self.handler.wait_for(*reply_method) as fut:
            self.send_method(method, (0, queue, no_ack))
            res = yield from fut
            if not isinstance(res, BasicMessage):
                raise QueueEmpty('Queue "{}" is empty!')
            return res

    def _basic_publish(self, message, exchange, routing_key, mandatory=False,
                       immediate=False, method=spec.BasicPublish):
        self.send_method(
            method,
            (0, exchange, routing_key, mandatory, immediate),
            message=message,
            frame_size=self.connection.frame_max,
        )
        # This should also be a generator
        yield

    # In case connection does not confirm publications
    basic_publish = _basic_publish

    def basic_publish_confirm(self, *args, reply_method=spec.BasicAck,
                              **kwargs):
        if not self._confirm_selected:
            self._confirm_selected = True
            self.confirm_select()
        with self.handler.wait_for(reply_method) as fut:
            self._basic_publish(*args, **kwargs)
            return (yield from fut)

    def basic_qos(self, prefetch_size, prefetch_count, a_global=False,
                  method=spec.BasicQos):
        self.send_method(method, (prefetch_size, prefetch_count, a_global))

    def basic_recover_async(self, requeue=False,
                            method=spec.BasicRecoverAsync):
        """Deprecated"""
        self.send_method(method, (requeue,))

    def basic_recover(self, requeue=False, method=spec.BasicRecover,
                      reply_method=spec.BasicRecoverOK):
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (requeue,))
            return (yield from fut)

    def basic_reject(self, delivery_tag, requeue=False,
                     method=spec.BasicReject):
        self.send_method(method, (delivery_tag, requeue))

    # Work with transactions

    def tx_select(self, method=spec.TxSelect, reply_method=spec.TxSelectOK):
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, ())
            return (yield from fut)

    def tx_commit(self, method=spec.TxCommit, reply_method=spec.TxCommitOK):
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, ())
            return (yield from fut)

    def tx_rollback(self, method=spec.TxRollback,
                    reply_method=spec.TxRollbackOK):
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, ())
            return (yield from fut)

    def confirm_select(self, nowait=False, method=spec.ConfirmSelect,
                       reply_method=spec.ConfirmSelectOK):
        reply_method = None if nowait else reply_method
        with self.handler.wait_for(reply_method) as fut:
            self.send_method(method, (nowait,))
            return (yield from fut)



# -*- coding: utf-8 -*-
from .serialization import (dumps, loads, SHORTSTR, TABLE, OCTET, TIMESTAMP,
                            SHORT, LONGLONG)

__all__ = ['BasicMessage']


def dump_properties(properties, types):
    flags = []
    flag_bits = 0
    shift = 15
    sformat, svalues = [], []
    get_prop = properties.get

    for name, proptype in types:
        value = get_prop(name)
        if value is not None:
            if shift == 0:
                # AMQP allows to have more than one set of properties
                flags.append(flag_bits)
                flag_bits = 0
                shift = 15

            flag_bits |= (1 << shift)
            sformat.append(proptype)
            svalues.append(value)
        shift -= 1
    return flag_bits.to_bytes(2, 'big') + dumps(''.join(sformat), svalues)


def load_properties(raw, types):
    (class_id, weight), offset = loads(SHORT+SHORT, raw)
    assert weight == 0, 'weight must always be 0'
    (body_size, flags), offset = loads(LONGLONG+SHORT, raw, offset=offset)

    flags = bin(flags)[2:]  # strip 0b
    sformat, snames = [], []
    for (name, proptype), flag in zip(types, flags):
        if flag == '1':
            sformat.append(proptype)
            snames.append(name)
    raw_properties, _ = loads(''.join(sformat), raw, offset=offset)
    return body_size, dict(zip(snames, raw_properties))


class BasicMessage(object):

    PROPERTIES = (
        ('content_type', SHORTSTR),
        ('content_encoding', SHORTSTR),
        ('headers', TABLE),
        ('delivery_mode', OCTET),
        ('priority', OCTET),
        ('correlation_id', SHORTSTR),
        ('reply_to', SHORTSTR),
        ('expiration', SHORTSTR),
        ('message_id', SHORTSTR),
        ('timestamp', TIMESTAMP),
        ('type', SHORTSTR),
        ('user_id', SHORTSTR),
        ('app_id', SHORTSTR),
    )

    delivery_info = None

    def __init__(self, body=None, **properties):
        properties.setdefault('content_encoding', 'utf-8')
        if isinstance(body, str):
            body = body.encode(properties['content_encoding'])
        self.body = [] if body is None else body
        self.properties = properties

        # Attributes below are used to create Message object from
        # sequence of frames
        self._body_len = None if body is None else len(body)
        self._body_received = 0
        self._ready = False

    def dump_properties(self, class_id):
        head = dumps('BBL', (class_id, 0, len(self.body)))
        return head + dump_properties(self.properties, self.PROPERTIES)

    def load_properties(self, raw, load_properties=load_properties):
        return load_properties(raw, self.PROPERTIES)

    @property
    def headers(self):
        return self.properties.get('headers')

    @property
    def delivery_tag(self):
        return self.delivery_info.get('delivery_tag')

    # Methods below are used to create Message object from
    # sequence of frames: header -> body [-> body ...]

    def add_header(self, raw):
        self._body_len, self.properties = self.load_properties(raw)
        self.properties.setdefault('content_encoding', 'utf-8')

    def add_body_chunk(self, chunk):
        if self._ready:
            return
        body_len = self._body_len
        assert body_len is not None

        self.body.append(chunk)
        self._body_received += len(chunk)
        assert body_len <= self._body_received, 'Received extra body chunk'

        if body_len == self._body_received:
            self.body = b''.join(self.body)
            self.body = self.body.decode(self.properties['content_encoding'])
            self._ready = True



# -*- coding: utf-8 -*-
"""
colibri.serialization
~~~~~~~~~~~~~~~~~~~~~

This module contains methods to convert between AMQP types
and their C-type representations.

"""
from __future__ import absolute_import

from io import BytesIO
from time import mktime
from decimal import Decimal
from datetime import datetime
from struct import pack, unpack_from

from .utils import maybe_list


# This may be confusing, but different broker implementations
# use type lables that differ from the spec
# Read more here: https://www.rabbitmq.com/amqp-0-9-1-errata.html#section_3
#
# Types below are defined for RabbitMq/QPID, if you have some issues
# using other broker please, report an issue on github
BIT = 'b'
OCTET = 'o'
SHORT = 'B'
LONG = 'l'
LONGLONG = 'L'
SHORTSTR = 's'
LONGSTR = 'S'
TABLE = 'F'
TIMESTAMP = 'L'

TYPE_BY_NAME = {
    'bit': BIT,
    'octet': OCTET,
    'short': SHORT,
    'long': LONG,
    'longlong': LONGLONG,
    'shortstr': SHORTSTR,
    'longstr': LONGSTR,
    'table': TABLE,
    'timestamp': TIMESTAMP,
}


def _read_item(buf, offset=0, unpack_from=unpack_from):
    ftype = chr(buf[offset])
    offset += 1

    # 'S': long string
    if ftype == 'S':
        slen, = unpack_from('!I', buf, offset)
        offset += 4
        val = buf[offset:offset + slen]
        if isinstance(val, bytes):
            val = val.decode('utf-8')
        offset += slen
    # 's': short string
    elif ftype == 's':
        slen, = unpack_from('!B', buf, offset)
        offset += 1
        val = buf[offset:offset + slen]
        if isinstance(val, bytes):
            val = val.decode('utf-8')
        offset += slen
    # 'b': short-short int
    elif ftype == 'b':
        val, = unpack_from('!B', buf, offset)
        offset += 1
    # 'B': short-short unsigned int
    elif ftype == 'B':
        val, = unpack_from('!b', buf, offset)
        offset += 1
    # 'U': short int
    elif ftype == 'U':
        val, = unpack_from('!h', buf, offset)
        offset += 2
    # 'u': short unsigned int
    elif ftype == 'u':
        val, = unpack_from('!H', buf, offset)
        offset += 2
    # 'I': long int
    elif ftype == 'I':
        val, = unpack_from('!i', buf, offset)
        offset += 4
    # 'i': long unsigned int
    elif ftype == 'i':
        val, = unpack_from('!I', buf, offset)
        offset += 4
    # 'L': long long int
    elif ftype == 'L':
        val, = unpack_from('!q', buf, offset)
        offset += 8
    # 'l': long long unsigned int
    elif ftype == 'l':
        val, = unpack_from('!Q', buf, offset)
        offset += 8
    # 'f': float
    elif ftype == 'f':
        val, = unpack_from('!f', buf, offset)
        offset += 4
    # 'd': double
    elif ftype == 'd':
        val, = unpack_from('!d', buf, offset)
        offset += 8
    # 'D': decimal
    elif ftype == 'D':
        d, = unpack_from('!B', buf, offset)
        offset += 1
        n, = unpack_from('!i', buf, offset)
        offset += 4
        val = Decimal(n) / Decimal(10 ** d)
    # 'F': table
    elif ftype == 'F':
        tlen, = unpack_from('!I', buf, offset)
        offset += 4
        limit = offset + tlen
        val = {}
        while offset < limit:
            keylen, = unpack_from('!B', buf, offset)
            offset += 1
            key = buf[offset:offset + keylen]
            if isinstance(key, bytes):
                key = key.decode('utf-8')
            offset += keylen
            val[key], offset = _read_item(buf, offset)
    # 'A': array
    elif ftype == 'A':
        alen, = unpack_from('!I', buf, offset)
        offset += 4
        limit = offset + alen
        val = []
        while offset < limit:
            v, offset = _read_item(buf, offset)
            val.append(v)
    # 't' (bool)
    elif ftype == 't':
        val, = unpack_from('!B', buf, offset)
        val = bool(val)
        offset += 1
    # 'T': timestamp
    elif ftype == 'T':
        val, = unpack_from('!Q', buf, offset)
        offset += 8
        val = datetime.utcfromtimestamp(val)
    # 'V': void
    elif ftype == 'V':
        val = None
    else:
        raise ValueError(
            'Unknown value in table: {0!r} ({1!r})'.format(
                ftype, type(ftype)))
    return val, offset


def loads(fmt, buf, offset=0, ord=ord, maybe_list=maybe_list,
          unpack_from=unpack_from, _read_item=_read_item):
    """
    bit = b
    octet = o
    short = B
    long = l
    long long = L
    float = f
    shortstr = s
    longstr = S
    table = F
    array = A
    """
    bitcount = bits = 0

    values = []
    append = values.append

    if len(fmt) == 1:
        buf = maybe_list(buf)

    for p in fmt:
        if p == 'b':
            if not bitcount:
                bits = ord(buf[offset:offset + 1])
            bitcount = 8
            val = (bits & 1) == 1
            bits >>= 1
            bitcount -= 1
            offset += 1
        elif p == 'o':
            bitcount = bits = 0
            val, = unpack_from('!B', buf, offset)
            offset += 1
        elif p == 'B':
            bitcount = bits = 0
            val, = unpack_from('!H', buf, offset)
            offset += 2
        elif p == 'l':
            bitcount = bits = 0
            val, = unpack_from('!I', buf, offset)
            offset += 4
        elif p == 'L':
            bitcount = bits = 0
            val, = unpack_from('!Q', buf, offset)
            offset += 8
        elif p == 'f':
            bitcount = bits = 0
            val, = unpack_from('!d', buf, offset)
            offset += 8
        elif p == 's':
            bitcount = bits = 0
            slen, = unpack_from('B', buf, offset)
            offset += 1
            val = buf[offset:offset + slen]
            if isinstance(val, bytes):
                val = val.decode('utf-8')
            offset += slen
        elif p == 'S':
            bitcount = bits = 0
            slen, = unpack_from('!I', buf, offset)
            offset += 4
            val = buf[offset:offset + slen]
            if isinstance(val, bytes):
                val = val.decode('utf-8')
            offset += slen
        elif p == 'F':
            bitcount = bits = 0
            tlen, = unpack_from('!I', buf, offset)
            offset += 4
            limit = offset + tlen
            val = {}
            while offset < limit:
                keylen, = unpack_from('!B', buf, offset)
                offset += 1
                key = buf[offset:offset + keylen]
                if isinstance(key, bytes):
                    key = key.decode('utf-8')
                offset += keylen
                val[key], offset = _read_item(buf, offset)
        elif p == 'A':
            bitcount = bits = 0
            alen, = unpack_from('!I', buf, offset)
            offset += 4
            limit = offset + alen
            val = []
            while offset < limit:
                aval, offset = _read_item(buf, offset)
                val.append(aval)
        elif p == 'T':
            bitcount = bits = 0
            val, = unpack_from('!Q', buf, offset)
            offset += 8
            val = datetime.fromtimestamp(val)
        else:
            raise Exception('WTF WTF')
        append(val)
    return values, offset


def _flushbits(bits, write, pack=pack):
    if bits:
        write(pack('B' * len(bits), *bits))
        bits[:] = []
    return 0


def dumps(fmt, values, maybe_list=maybe_list):
    """"
    bit = b
    octet = o
    short = B
    long = l
    long long = L
    shortstr = s
    longstr = S
    table = F
    array = A
    """
    bitcount = 0
    bits = []
    out = BytesIO()
    write = out.write

    if len(fmt) == 1:
        values = maybe_list(values)

    for i, val in enumerate(values):
        p = fmt[i]
        if p == 'b':
            val = 1 if val else 0
            shift = bitcount % 8
            if shift == 0:
                bits.append(0)
            bits[-1] |= (val << shift)
            bitcount += 1
        elif p == 'o':
            bitcount = _flushbits(bits, write)
            write(pack('B', val))
        elif p == 'B':
            bitcount = _flushbits(bits, write)
            write(pack('!H', int(val)))
        elif p == 'l':
            bitcount = _flushbits(bits, write)
            write(pack('!I', val))
        elif p == 'L':
            bitcount = _flushbits(bits, write)
            write(pack('!Q', val))
        elif p == 's':
            val = val or ''
            bitcount = _flushbits(bits, write)
            write(pack('B', len(val)))
            if isinstance(val, str):
                val = val.encode('utf-8')
            write(val)
        elif p == 'S':
            val = val or ''
            bitcount = _flushbits(bits, write)
            write(pack('!I', len(val)))
            if isinstance(val, str):
                val = val.encode('utf-8')
            write(val)
        elif p == 'F':
            bitcount = _flushbits(bits, write)
            _write_table(val or {}, write, bits)
        elif p == 'A':
            bitcount = _flushbits(bits, write)
            _write_array(val or [], write, bits)
        elif p == 'T':
            write(pack('!q', int(mktime(val.timetuple()))))
    _flushbits(bits, write)

    return out.getvalue()


def _write_table(d, write, bits, pack=pack):
    out = BytesIO()
    twrite = out.write
    for k, v in d.items():
        twrite(pack('B', len(k)))
        if isinstance(k, str):
            k = k.encode('utf-8')
        twrite(k)
        try:
            _write_item(v, twrite, bits)
        except ValueError:
            raise ValueError('Table type {!r} for key {!r} not handled by '
                             'amqp.[value: {!r}]'.format(type(v), k, v))
    table_data = out.getvalue()
    write(pack('!I', len(table_data)))
    write(table_data)


def _write_array(l, write, bits, pack=pack):
    out = BytesIO()
    awrite = out.write
    for v in l:
        try:
            _write_item(v, awrite, bits)
        except ValueError:
            raise ValueError('Table type {0!r} not handled by amqp. '
                             '[value: {1!r}]'.format(type(v), v))
    array_data = out.getvalue()
    write(pack('!I', len(array_data)))
    write(array_data)


def _write_item(v, write, bits, pack=pack,
                str=str, bytes=bytes, bool=bool,
                float=float, int=int,
                Decimal=Decimal, datetime=datetime, dict=dict, list=list,
                tuple=tuple, None_t=None):
    if isinstance(v, str):
        write(pack('!cI', b'S', len(v)))
        write(v.encode('utf-8'))
    elif isinstance(v, bytes):
        write(pack('!cI', b'S', len(v)))
        write(v)
    elif isinstance(v, bool):
        write(pack('!cB', b't', int(v)))
    elif isinstance(v, float):
        write(pack('!cd', b'd', v))
    elif isinstance(v, int):
        write(pack('!ci', b'I', v))
    elif isinstance(v, Decimal):
        sign, digits, exponent = v.as_tuple()
        v = 0
        for d in digits:
            v = (v * 10) + d
        if sign:
            v = -v
        write('!cBi', b'D', -exponent, v)
    elif isinstance(v, datetime):
        # TODO use UTC
        write(pack('!cq', b'T', int(mktime(v.timetuple()))))
    elif isinstance(v, dict):
        write(b'F')
        _write_table(v, write, bits)
    elif isinstance(v, (list, tuple)):
        write(b'A')
        _write_array(v, write, bits)
    elif v is None_t:
        write(b'V')
    else:
        raise ValueError



# -*- coding: utf-8 -*-
"""
colibri.utils
~~~~~~~~~~~~~

Various useful classes and functions.

"""
from __future__ import absolute_import

from asyncio import async
from collections import Iterable, Mapping


class cached_property(object):
    """Property descriptor that caches the return value
    of the get function.

    *Examples*

    .. code-block:: python

        @cached_property
        def connection(self):
            return Connection()
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls=None):
        r = instance.__dict__[self.func.__name__] = self.func(instance)
        return r


def is_list(obj, scalars=(Mapping, str), iterables=(Iterable, )):
    """Return True if the object is iterable, but not if object is one
    of `scalars`."""
    return isinstance(obj, iterables) and not isinstance(obj, scalars or ())


def maybe_list(obj, scalars=(Mapping, str)):
    """Return list of one element if `obj` is a scalar."""
    return obj if obj is None or is_list(obj, scalars) else [obj]


def chunks(iterable, chunk_size):
    """Yield successive `chunk_size`d chunks from `iterable`."""
    for i in range(0, len(iterable), chunk_size):
        yield iterable[i:i+chunk_size]


class YieldFromContextManager(object):
    """Context manager.

    This enables the following idiom for opening and closing an
    object around a block:

        with (yield from obj):
            <block>

    while failing loudly when accidentally using:

        with obj:
            <block>
    """

    def __init__(self, obj):
        self._obj = obj

    def __enter__(self):
        return self._obj

    def __exit__(self, *args):
        try:
            # contextmanager does not yield from __exit__
            async(self._obj.close())
        finally:
            self._obj = None  # Crudely prevent reuse



# -*- coding: utf-8 -*-
from .frame import HeartbeatFrame
from .spec import ConnectionClose, FRAME_ERROR

# TODO fix this


class HeartbeatMonitor(object):

    frame = HeartbeatFrame.pack()

    def __init__(self, protocol, interval, timeout=None):
        self.protocol = protocol
        self.interval = interval
        self.timeout = timeout or interval * 2
        self.on_timeout = None

    def start(self, interval=None, timeout=None):
        self.send(interval)
        self.monitor(timeout)

    def send(self, interval=None):
        if interval:
            self.interval = interval
            self.monitor = interval * 2
        protocol = self.protocol
        protocol._send_frame(self.frame)
        protocol.loop.call_later(self.interval, self.send)

    def monitor(self, timeout=None):
        if timeout:
            self.timeout = timeout
        self.on_timeout = self.protocol.loop.call_later(self.timeout,
                                                        self.timed_out)

    def timed_out(self, FRAME_ERROR=FRAME_ERROR):
        self.protocol.send_method(
            0, ConnectionClose,
            (FRAME_ERROR, 'Heartbeat timed out', 0, 0),
        )

    def maybe_received(self):
        if self.on_timeout is not None:
            self.on_timeout.cancel()
            self.monitor()



# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup, find_packages


if sys.version_info < (3, 3):
    raise Exception('colibri requires Python 3.3 or higher')


# -*- Distribution Meta -*-
meta = {}
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'colibri/_version.py')) as f:
    for line in f:
        key, value = line.split('=')
        meta[key.strip()] = value.strip("' \n")


with open('README.rst') as f:
    README = f.read()


def strip_comments(l):
    return l.split('#', 1)[0].strip()


def reqs(*f):
    return [
        r for r in (
            strip_comments(l) for l in open(
                os.path.join(os.getcwd(), 'requirements', *f)).readlines()
        ) if r]


install_requires = reqs('default.txt')
if sys.version_info < (3, 4):
    install_requires.extend(reqs('py3.txt'))


setup(
    name='colibri',
    version=meta['__version__'],
    author=meta['__author__'],
    author_email=['__contact__'],
    description='asyncio-based implementation of AMQP client',
    long_description=README,
    keywords='colibri AMQP client asyncio',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    url='http://pypi.python.org/pypi/colibri',
    license='MIT',
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=install_requires,
    include_package_data=False,
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""
colibri.exceptions
~~~~~~~~~~~~~~~~~~

This module contains colibri-specific exceptions.

"""
from __future__ import absolute_import


class ColibriException(Exception):
    """Base class for all exceptions."""


class UnexpectedFrame(ColibriException):
    """Raised when unexpected frame is received"""


class QueueEmpty(ColibriException):
    """Raised when trying to BasicGet from empty queue"""
